/* netio.cpp - handle IO from the net
 *
 * Copyright (c) 1997-2002 Tony Aiuto
 *
 * To report problems with this tool or to suggest enhancements send mail
 * to tony@aiu.to
 *
 */

#ifndef lint
static char _copyright[] = "@(#) Copyright (c) 1997-2002, Tony Aiuto";
#endif

/*
 * Author: Tony Aiuto (tony@aiu.to)
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor their employers are responsible for
 *    any consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors must appear
 *    in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This copyright notice may not be removed or altered.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL TONY AIUTO, HIS CURRENT OR PAST EMPLOYERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Edit History:
 *
 * $Log: netio.cpp $
 * Revision tools_TonyAiuto/4 2004/02/23 18:46:14 TonyAiuto
 * 	5.4 release
 * 
 * Revision tools_TonyAiuto/3 2004/01/07 17:51:30 TonyAiuto
 * 	5.3.4 savepoint after HPUX and windows initial tweaks
 * 
 * Revision tools_TonyAiuto/2 2003/11/21 03:46:14 TonyAiuto
 * 	changes for 5.3.0
 * 
 */

#ifndef lint
static char rcsid[] = "@(#) $Id: netio.cpp tools_TonyAiuto/4 2004/02/23 18:46:14 TonyAiuto $";
#endif

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include <stdlib.h>
#include <assert.h>

#include <sys/types.h>
#include <sys/stat.h>

// Bring in the platform specific indcludes

#if defined(WIN32)

#include <winsock2.h>
#include <shlwapi.h>
#define sleep(x) _sleep(x)

#else

#include <unistd.h>
#include <sys/socket.h>
#include <sys/utsname.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#endif

#include <signal.h>

#if defined(WIN32)
static void alarm(int x) { }
#define	SIGALRM	-1234
#endif



#if HAVE_SSL > 0
#include <sslfuncs.h>
#endif
#include "netio.h"


extern	int	verbose;


NetIOStream::NetIOStream(int _ios)
{
	ios = _ios;
	idx = 0;
	bufcount = 0;
	timeout = 60;
}

NetIOStream::~NetIOStream()
{
}

void NetIOStream::setTimeout(int _timeout)
{
	timeout = _timeout;
}

int NetIOStream::fill()
{
	int	null = 0;

	idx = 0;
	while(null < 3) {
		alarm(timeout);
		bufcount = ::recvfrom(ios, (char*) b, sizeof(b), 0, NULL, NULL);
		if(bufcount != 0) {
			// printf("filled %d bytes\n", bufcount);
			return bufcount;
		}

		// count empty reads
		null++;
		// printf("Got null %d\n", null);
	}
	return -1;
}


int NetIOStream::nextByte()
{
	if(idx >= bufcount) {
		if(fill() < 0) return -1;
	}
	return (int) b[idx++];
}

int NetIOStream::readLine(char *buf, int bufsize)
{
	int 	c;
	int	nbytes = 0;
	int	lastWasCR = 0;
	char	*ret = buf;

	while(bufsize > 1) {
		c = nextByte();
		if(c < 0) break;

		*buf++ = c;
		nbytes += 1;
		bufsize -= 1;

		if(lastWasCR && (c == '\n') ) break;
		if(c == '\n') break;

		if(c == '\r') {
			lastWasCR = 1;
		} else {
			lastWasCR = 0;
		}
	}
	*buf = '\0';
	// printf("readLine: %s", ret);
	return nbytes;
}


int NetIOStream::read(char *buf, int bufsize)
{
	int	null = 0;

	while(idx >= bufcount) {
		if(fill() < 0) return -1;
	}

	int	nleft = bufcount - idx;
	if(bufsize > nleft) bufsize = nleft;
	memcpy(buf, &b[idx], bufsize);
	idx += bufsize;
	return bufsize;
}


/* send data over the network
 * We can't just write big gobs, so we break it up and 
 * make sure network throttleing does not stop us
 */

int NetIOStream::send(const char *buf, size_t bufsize)
{
	int ret = 0;
	int sent = 0;
	while(bufsize > 0) {
		size_t nsend = 1024;
		if(bufsize < nsend) nsend = bufsize;
		// printf("writing %d bytes\n", nsend);
#if defined(MSG_DONTWAIT)
		int status = ::send(ios, buf, nsend, MSG_DONTWAIT);
#else
		int status = ::send(ios, buf, nsend, 0);
#endif
		// printf("ret = %d\n", status);
		if(status > 0) {
			buf += status;
			sent += status;
			bufsize -= status;
		} else {
			int err = errno;
			// printf("EAGAIN = %d, status = %d: %s\n", EAGAIN, err, strerror(err));
#if defined(EWOULDBLOCK)
			if (err == EAGAIN || err == EWOULDBLOCK) {
#else
			if (err == EAGAIN) {
#endif
				// fprintf(stderr, "sleeping\n");
				sleep(1);
			} else {
				errno = err;
				return status;
			}
		}
	}
	return sent;
}

#if HAVE_SSL > 0

NetIOStreamSSL::NetIOStreamSSL(SimpleSSL *_ssl, int _ios) :
	NetIOStream(_ios),
	ssl(_ssl)
{
	//printf("NetIOStreamSSL - ctor()\n");
}

NetIOStreamSSL::~NetIOStreamSSL()
{
}

int NetIOStreamSSL::fill()
{
	int	null = 0;

	idx = 0;
	while(null < 3) {
		alarm(timeout);

		bufcount = ssl->read((char*) b, sizeof(b));
		if(bufcount != 0) {
			// printf("filled %d bytes\n", bufcount);
			return bufcount;
		}

		// count empty reads
		null++;
		// printf("Got null %d\n", null);
	}
	return -1;
}

int NetIOStreamSSL::send(const char *buf, size_t bufsize)
{
	return ssl->write(buf, bufsize);
}

#endif
