/* sslfuncs.c - ssl support functions
 *
 * Copyright (c) 1997-2004 Tony Aiuto
 *
 * To report problems with this tool or to suggest enhancements send mail
 * to tony@aiu.to
 *
 */

#ifndef lint
static char _copyright[] = "@(#) Copyright (c) 1997-2004, Tony Aiuto";
#endif

/*
 * Author: Tony Aiuto (tony@aiu.to)
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor their employers are responsible for
 *    any consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors must appear
 *    in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This copyright notice may not be removed or altered.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL TONY AIUTO, HIS CURRENT OR PAST EMPLOYERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <assert.h>
#include <errno.h>
#include <string.h>


#include <openssl/bio.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/rand.h>

#include "sslfuncs.h"
#include "ssloptions.h"

extern int verbose;		// global from main.cpp

SimpleSSL *SimpleSSL::singleton = NULL;
int SimpleSSL::dataIndex = -1;

// module global pointer to any options
static SSLOptions *options = NULL;

/* returns the number of printed errors */
int SimpleSSL::printErrors() 
{
	int	nerr = 0;
	unsigned long curerr = 0;
	char errbuff[1024];
	memset(errbuff, 0, sizeof(errbuff));
	for (curerr = ERR_get_error (); curerr; curerr = ERR_get_error ()) {
		fprintf(stderr, "SSL: %s\n", ERR_error_string(curerr, errbuff));
		++nerr;
	}
  	return nerr;
}


/** setSSLOptions - associates SSL options with this instance.
 * The caller still owns the data and is expected to not destroy
 * the SSLOptions instance until after the SimpleSSL library is
 * no longer needed
 */
void SimpleSSL::setOptions(SSLOptions *_ssloptions)
{
        options = _ssloptions;
}


/* foundary to make sure we only create one of these */
SimpleSSL *SimpleSSL::getSimpleSSL()
{
	static	int	initialized = 0;

	if(singleton != NULL) return singleton;

	if(initialized == 0) {
		// Bring up library
		SSL_load_error_strings();
		SSL_library_init();
		initRandom();
		initialized = 1;
		// allocate a new index to store things in the SSL struct
		dataIndex = SSL_get_ex_new_index(0, NULL, NULL, NULL, NULL);
	}

	singleton = new SimpleSSL();
	if(singleton->getContext() == NULL) {
		fprintf(stderr, "Could not get SSL Context\n");
		return NULL;
	}
	return singleton;
}


/* protected ctor.   We don't want a bunch of these floating around
 */

SimpleSSL::SimpleSSL()
{
	int err = init();
	if(err != 0) {
		fprintf(stderr, "Could not initialize SSL library, err=%d.\n", err);
	}
}

SimpleSSL::~SimpleSSL()
{
	SSL_CTX_free(sslContext);
	sslContext = NULL;
}


void SimpleSSL::initRandom()
{
	char randFile[256];
	int maxrand = 500;

	// pick up default random # seed file name
	RAND_file_name(randFile, sizeof (randFile));
	if (&randFile[0]) {
		// Seed at most 16k (value borrowed from wget)
		RAND_load_file(randFile, 16384);
	}

	if (RAND_status()) return;

	if (RAND_status()) return;

#if defined(WIN32)
		// Under Windows, we can try to seed the PRNG using
		// screen content. This may or may not work

		RAND_screen();
		if (RAND_status()) return;
#endif

  /* Still not enough randomness, presumably because neither random
     file nor EGD have been available.  Use the stupidest possible
     method -- seed OpenSSLs PRNG with the systems PRNG.  This is
     insecure in the cryptographic sense, but people who care about
     security will use /dev/random or their own source of randomness
     anyway.  */

#if !defined(WIN32)
	while (RAND_status() == 0 && maxrand-- > 0) {
		unsigned char rnd = (unsigned char) random();
		RAND_seed (&rnd, sizeof (rnd));
	}
#endif

	if (RAND_status () == 0) {
		fprintf(stderr, "Could not initialize SSL random #s\n");
	}
}



/* Creates a SSL Context and sets some defaults for it
 * Returns: 0 for success
 * Side Effects:
 *	sets sslContext on success
 */

int SimpleSSL::init()
{
	SSL_METHOD *meth = NULL;
	int verifyMode = SSL_VERIFY_NONE;

	myConnection = NULL;

	// make the context
	meth = SSLv23_client_method();
	sslContext = SSL_CTX_new(meth);
	// printf("sslContext = 0x%x\n", sslContext);
	if (sslContext == NULL) return SimpleSSL_COULD_NOT_CREATE_CONTEXT;

	SSL_CTX_set_verify_depth(sslContext, 9);
	if(options != NULL) {
		if(options->getCertFile() != NULL ||
		   options->getCertPath() != NULL) {
			if(SSL_CTX_load_verify_locations(sslContext,
							options->getCertFile(),
							options->getCertPath()
								) != 1) {
				printErrors();
				return SimpleSSL_BAD_CERT_PATHS;
			}
		}
		const char *clientKeyFile = options->getCertKeyFile();
		if(clientKeyFile != NULL && *clientKeyFile != '\0') {
			// I am not sure this is correct (tony 2003/12/29)
			if(SSL_CTX_use_certificate_chain_file(sslContext,
					options->getCertKeyFile()) <= 0) {
				printErrors();
				return SimpleSSL_INVALID_CLIENT_CERTIFICATE;
			}

			if(SSL_CTX_use_PrivateKey_file(sslContext, clientKeyFile,
						SSL_FILETYPE_PEM) <= 0) {
				printErrors();
				return SimpleSSL_INVALID_CLIENT_CERTIFICATE;
			}
		}
	}
	SSL_CTX_set_verify(sslContext, verifyMode, verifyCallback);

	return 0; /* Succeded */
}


/// see if the certificate verification error is allowable
/// according to our option set

static int isAllowedProblem(int err, int depth)
{
	int	ret = 0;
	switch (err) {
	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
		if(options != NULL && options->getAllowSelfSigned()) {
			ret = 1;
		}
		break;

	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
	case X509_V_ERR_CERT_UNTRUSTED:
	case X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE:
		if(options != NULL && options->getAllowUntrusted()) {
			ret = 1;
		}
		break;

	case X509_V_ERR_CERT_HAS_EXPIRED:
		if(options != NULL && options->getAllowExpired()) {
			ret = 1;
		}

	case X509_V_ERR_CERT_NOT_YET_VALID:
	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
		break;

	}
	return ret;
}



/* TODO - make this private static in base class and call 
 * virtualized one
 */

int SimpleSSL::verifyCallback(int ok, X509_STORE_CTX *ctx)
{
	char *s, buf[256];
	X509	*err_cert;
	int	err, depth;
	SSL	*ssl;
	SimpleSSL *self;

	err_cert = X509_STORE_CTX_get_current_cert(ctx);
	err = X509_STORE_CTX_get_error(ctx);
	depth = X509_STORE_CTX_get_error_depth(ctx);

	if(verbose) fprintf(stderr, "SSL::verifyCB: ok=%d, err=%d, depth=%d\n", ok, err, depth);

	/*
	 * get back the pointer to the SimpleSSL which made us.
	 * - first get the global index in the X509_STORE_CTX which
	 * tells us where the SSL* is stored
	 * - then get the SSL* from that
	 * - then get ourself back from the SSL struct
	 */
	int idx =  SSL_get_ex_data_X509_STORE_CTX_idx();
	ssl = (SSL *) X509_STORE_CTX_get_ex_data(ctx, idx);
	self = (SimpleSSL *) SSL_get_ex_data(ssl, dataIndex);
	if(verbose) fprintf(stderr, "SSL::verifyCB: got self @ 0x%x\n", self);

	s = X509_NAME_oneline(X509_get_subject_name(err_cert), buf, 256);
	if(verbose) fprintf(stderr, "X509 Name=%s\n", s);

	if (ok == 0) {
		ok = isAllowedProblem(err, depth);
		if(ok == 1) {
			if(verbose) printf("verify set OK = 1\n");
		}
	}
	return ok;
}



/* Sets up a SSL structure and performs the handshake on fd 
 * @return: 0 for success, else error code
 */

int SimpleSSL::connect(int fd)
{
	return connect(myConnection, fd);
}

int SimpleSSL::connect(SSL *&con, int fd)
{
	con = SSL_new(sslContext);
	SSL_set_verify_depth(con, 9);
	// printf("sslContext = 0x%x\n", sslContext);
	// printf("ssl connection = 0x%0x\n", con);

	// save a pointer to myself in the userdata section of con.
	// This is needed for the verifyCallback
	if(verbose) fprintf(stderr, "storing self 0x%x at slot %d\n", this, dataIndex);
	SSL_set_ex_data(con, dataIndex, (void *) this);

	if(SSL_set_fd(con, fd) == 1) {
		SSL_set_connect_state(con); 
		int err =  SSL_connect(con);  

		if(err != 1) {
			printErrors();
#if TODO
			int eret = SSL_get_error(con, err);

			switch (eret) {
			case SSL_ERROR_NONE:
				printf("no error\n");
				break;
			case SSL_ERROR_SSL:
				printErrors();
				break;
			case SSL_ERROR_ZERO_RETURN:
				printf("SSL_ERROR_ZERO_RETURN\n");
				break;
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_WRITE:
				printf("SSL_ERROR_WANT_READ/WRITE\n");
				break;

			case SSL_ERROR_WANT_CONNECT:
				printf("SSL_ERROR_WANT_CONNECT\n");
				break;
			default:
				printf("Got SSLerror %d, %x\n", eret, eret);
			}
#endif
		} else if (con->state != SSL_ST_OK) {
printf("SimpleSSL::connect: state = %d\n", (con)->state);
			return 1;
		}

/*
#define SSL_get_state(a)                SSL_state(a)
#define SSL_is_init_finished(a)         (SSL_state(a) == SSL_ST_OK)
#define SSL_in_init(a)                  (SSL_state(a)&SSL_ST_INIT)
#define SSL_in_before(a)                (SSL_state(a)&SSL_ST_BEFORE)
#define SSL_in_connect_init(a)          (SSL_state(a)&SSL_ST_CONNECT)
#define SSL_in_accept_init(a)           (SSL_state(a)&SSL_ST_ACCEPT)
*/
	}

	int SSLerror;
	while((SSLerror = ERR_get_error())!=0)
		printf("%s\n", ERR_error_string(SSLerror,NULL));

	int err = SSL_get_verify_result(con);
	if(err != X509_V_OK) {
		if(verbose) fprintf(stderr, "SSL failure => %d\n", err);
		err = isAllowedProblem(err, 0);
		if(verbose) fprintf(stderr, "isAllowed => %d\n", err);
		return err == 0;
	}
	return 0;
}


#ifdef notdef
void
shutdown_ssl (SSL* con)
{
  SSL_shutdown (con);
  if (con != NULL)
    SSL_free (con);
}
#endif


/* SSL version of iread. Only exchanged read for SSL_read 
   Read at most LEN bytes from FD, storing them to BUF.  This is
   virtually the same as read(), but takes care of EINTR braindamage
   and uses select() to timeout the stale connections (a connection is
   stale if more than OPT.TIMEOUT time is spent in select() or
   read()).  */
int SimpleSSL::read(char *buf, int len)
{
	return read(myConnection, buf, len);
}

int SimpleSSL::read(SSL *con, char *buf, int len)
{
	int res;
	int fd;

	BIO_get_fd(con->rbio, &fd);

  do {
#ifdef HAVE_SELECT
printf("Doing select\n");
      if (opt.timeout && !SSL_pending (con))
	{
	  do
	    {
	      res = select_fd (fd, opt.timeout, 0);
	    }
	  while (res == -1 && errno == EINTR);
	  if (res <= 0)
	    {
	      /* Set errno to ETIMEDOUT on timeout.  */
	      if (res == 0)
		/* #### Potentially evil!  */
		errno = ETIMEDOUT;
	      return -1;
	    }
	}
#endif
      res = SSL_read (con, buf, len);
    }
  while (res == -1 && errno == EINTR);

  return res;
}


/* SSL version of iwrite. Only exchanged write for SSL_write 
   Write LEN bytes from BUF to FD.  This is similar to iread(), but
   doesn't bother with select().  Unlike iread(), it makes sure that
   all of BUF is actually written to FD, so callers needn't bother
   with checking that the return value equals to LEN.  Instead, you
   should simply check for -1.  */
int SimpleSSL::write(const char *buf, int len)
{
	return write(myConnection, buf, len);
}

int SimpleSSL::write(SSL *con, const char *buf, int len)
{
	int res = 0;
	int fd;
	BIO_get_fd (con->rbio, &fd);

  /* `write' may write less than LEN bytes, thus the outward loop
     keeps trying it until all was written, or an error occurred.  The
     inner loop is reserved for the usual EINTR f*kage, and the
     innermost loop deals with the same during select().  */

	while (len > 0) {
		do {
#ifdef HAVE_SELECT
			if (opt.timeout) {
				do {
					res = select_fd (fd, opt.timeout, 1);
				} while (res == -1 && errno == EINTR);
				if (res <= 0) {
					/* Set errno to ETIMEDOUT on timeout. */
					if (res == 0) {
						/* #### Potentially evil!  */
						errno = ETIMEDOUT;
					}
					return -1;
				}
			}
#endif
			res = SSL_write (con, buf, len);
		} while (res == -1 && errno == EINTR);
		if (res <= 0) break;
		buf += res;
		len -= res;
	}
	return res;
}

#if OPENSSL_VERSION_NUMBER > 0x00907000
#endif
