/* httpfetch.h
 *
 */

/* Edit History:
 * $Log: webfetch.h $
 * Revision tools_TonyAiuto/15 2004/05/28 15:36:24 TonyAiuto
 * 	5.4.1 checkpoint
 * 
 * Revision tools_TonyAiuto/14 2004/02/23 18:46:14 TonyAiuto
 * 	5.4 release
 * 
 * Revision tools_TonyAiuto/13 2004/01/07 17:51:30 TonyAiuto
 * 	5.3.4 savepoint after HPUX and windows initial tweaks
 * 
 * Revision tools_TonyAiuto/12 2003/11/21 03:46:14 TonyAiuto
 * 	changes for 5.3.0
 * 
 * Revision tools_TonyAiuto/11 2003/11/15 01:36:23 TonyAiuto
 * 	SSL working, but cert verification flakey
 * 
 * Revision tools_TonyAiuto/10 2003/11/13 20:30:19 TonyAiuto
 * 	interim checkpoint
 * 
 * Revision tools_TonyAiuto/9 2003/10/26 03:15:18 TonyAiuto
 * 	5.2.1 - move placement of tobase64
 * 
 * Revision tools_TonyAiuto/8 2003/08/21 14:33:55 TonyAiuto
 * 	5.2.0 cookie handling changes
 * 
 * Revision tools_TonyAiuto/7 2002/10/17 18:08:22 TonyAiuto
 * 	5.1.0 release
 * 
 * Revision tools_TonyAiuto/6 2002/08/07 19:58:25 TonyAiuto
 * 	5.0.0 beta
 * 
 * Revision tools_TonyAiuto/5 2002/07/18 16:01:36 TonyAiuto
 * 	initial 5.0.0 work
 * 
 */

#ifndef __httpfetch_h__
#define __httpfetch_h__

#include <cookie.h>
#include <proxy.h>

/* commands for HTTPop */

#define CMD_GET         0
#define CMD_POST        1
#define CMD_DELETE      2
#define CMD_HEAD        3
#define CMD_PUT		4

#define CMD_PROPFIND	5
#define CMD_PROPDEL	6
#define CMD_REPORT	7

class Webfetch;
class UploadData;
class SSLOptions;

typedef void (HeaderHandler)(Webfetch *, const char *buf, int len);


/** DataConsumer is an abstract class representing a sink
 * for a stream of retreived data.
 * Known subclasses are
 *	PrintingDataConsumer, which emits data to stdout
 *	FileDataConsumer, which saves data to a file
 */

class  DataConsumer {
public:
	DataConsumer();
	virtual ~DataConsumer();

	virtual void startStream(const char *url) = 0;
	virtual void takeData(const char *, int) = 0;
	virtual void endStream() = 0;
};

class PrintingDataConsumer : public DataConsumer {
public:
	PrintingDataConsumer();
	~PrintingDataConsumer();

	void startStream(const char *url);
	void takeData(const char *, int);
	void endStream();
};

class FileDataConsumer : public DataConsumer {
public:
	FileDataConsumer();
	~FileDataConsumer();

	void startStream(const char *url);
	void takeData(const char *, int);
	void endStream();

	void setOutputFile(const char *filename);
	void setCreateDirectories(int);

private:
	int	createDir;	// create directories needed to save file
	char	*outputFileName;
	FILE	*ofile;
};





class Protocol_Handler;

class Webfetch {

public:
	Webfetch();
	~Webfetch();

	void shutdown();
	void processURL(int command, const char *url, const UploadData *);


	///
	/// methods which set protocol behavior
	///

	void setAuthorization(const char *);
	const char *getAuthorization() { return authorization; }
	void setCookie(const char *);
	const char *getCookie() { return cookie; }
	void setConnectHost(const char *hostname);
	void setContentType(const char *contentType);
	const char *getContentType() { return contentType; }
	void setFollowRedirect(int);
	int getFollowRedirect() { return followRedirect; }
	void setFrom(const char *);
	const char *getFrom() { return from; }
	void setLanguage(const char *);
	const char *getLanguage() { return language; }
	void setProxyAuthorization(const char *a) { proxies.setAuthorization(a); }
	const char *getProxyAuthorization() { return proxies.getAuthorization(); }
	void setReferer(const char *);
	const char *getReferer() { return referer; }
	void setTimeout(long);
	long getTimeout() { return timeout; }
	const char *getUserAgent() { return userAgent; }
	void setUserAgent(const char *);
	void setVirtualHost(const char *hostname);	// deprecated

	///
	/// methods which configure inputs and outputs
	///

	void setDataConsumer(DataConsumer *);
	DataConsumer *getDataConsumer() { return consumer; }
	/* add a handler for header data.  The better way is
	 * to subclass and implement handleHeader, but this is
	 * a convenience for quick and dirty programs
	 */
	void	setHeaderHandler(HeaderHandler *);
	void setCookieJar(CookieJar *);
	CookieJar *getCookieJar() { return cookieJar; }
#if HAVE_SSL > 0
	void setSSLOptions(SSLOptions *);
#endif

	///
	/// methods which configure inputs and outputs
	///

	int	getHTTPStatus() { return httpStatus; }
	char	*getHTTPMessage() { return httpMessage; }
	int	getErrorStatus() { return errorStatus; }

	/* resolve hostname and port to sock structures, but deal
	 * with proxy stuff as well
	 */
	int resolveHosts(const char *host, int port,
		struct sockaddr_in *hostIP, struct sockaddr_in *connectIP);


	///
	/// utility functions
	///

	/** encode a string in base64 */
	static char *tobase64(const char *);

	///
	/// Methods that protocol handlers need, but are not really public

	virtual void handleHeader(const char *, int);	// header data callback
	int negotiateWithSocks4Proxy(int proxySocket, struct sockaddr_in *host);

	///
	/// Data
	///
public:

	// error codes
	enum { UNKNOWN_HOST=1, UNSUPPORTED_SCHEME=2, BAD_HOSTNAME=3,
		BAD_SOCKS_PROXY_HOST=4, BAD_HTTP_PROXY_HOST=5,
		CONNECTION_REFUSED=6,

		SSL_CERTIFICATE_FAILURE=10,

		OTHER_FAILURE=99
	};


	ProxyInfo	proxies;


	char	*virtualHost;	// virtual host to connect to (deprecated)


protected:


	// parameters about our personality

	char	*authorization;	// HTTP-Basic authorization 
	char	*cookie;	// extra cookie to send along
	CookieJar	*cookieJar;
	char	*connectHost;	// physical host to connect to
	int	followRedirect;	// follow Location headers
	int	errorStatus;	// error indicators if fetch failed
	char   	*from;		// content of From header
	char   	*language;	// content of Language header
	char   	*referer;	// referrer to spoof
	char	*contentType;	// Content-Type to send for POST data
	long	timeout;
	char   	*userAgent;	// user agent to spoof

	HeaderHandler	*headerHandler;
	DataConsumer	*consumer;

	// transient stuff for getting the file
	int	httpStatus;
	char	*httpMessage;
	Protocol_Handler	*current_handler;	// only for unexpected shutdown


};


/** Protocol_Handler is an abstract class which handles
 * communications to a server using a particular protocol
 * Known subclasses are
 *	HTTP_Handler - for handling HTTP
 *	HTTPS_Handler - for handling HTTPS
 */

class NetIOStream;

class Protocol_Handler {
public:
	Protocol_Handler(Webfetch *, const char *, int);
	virtual ~Protocol_Handler();

	virtual void doit(int cmd, const char *resource, const UploadData *args) = 0;
	virtual void processHeader(const char *, int) = 0;  // header interpretation

	/** establish connection to a physical server.
	 * @return status (0 for OK)
	 * Side Effect: sets serverFD
	 */
	int TCPconnect(const char *host);
	virtual int getNetIOStream();
	void shutdown();

	int	httpStatus;

protected:

public:

protected:

	Webfetch *context;		/* overall settings */
	char	*host;			/* hostname from url */
	int	port;			/* port to connect to */

	// transient stuff for getting the file
	int	serverFD;		// FD of remote host
	NetIOStream *hostio;		// actuall thing to talk to host

private:
};

class HTTP_Handler : public Protocol_Handler {
public:
	HTTP_Handler(Webfetch *, const char *, int);
	~HTTP_Handler();

	void doit(int cmd, const char *resource, const UploadData *args);
	void processHeader(const char *, int);

protected:
	// transient stuff for getting the file
	int	chunked;
	long	contentLength;
	char	*httpMessage;
	char	*redirect;
};


class HTTPS_Handler : public HTTP_Handler {
public:
	HTTPS_Handler(Webfetch *, const char *, int);
	~HTTPS_Handler();

protected:
	int getNetIOStream();
};



/* UploadData handles data to be sent in a POST or PUT request.
 *
 * It is responsible for holding the aspects about itself and
 * then for being able to stream it to the server.
 *
 */
 
class UploadData {
protected:
	UploadData();
public:
	virtual ~UploadData();
	virtual int getSize() const = 0;
	virtual int send(NetIOStream *) const = 0;
	virtual const char *getContentType() const;
};

class StringUploadData : public UploadData {
public:
	StringUploadData(const char *data);
	~StringUploadData();
	int getSize() const;
	int send(NetIOStream *) const;
	// const char *getContentType() const;	use default

protected:
	char	*sdata;
};

class FileUploadData : public UploadData {
public:
	FileUploadData(const char *filename);
	~FileUploadData();
	int getSize() const;
	int send(NetIOStream *) const;
	const char *getContentType() const;

protected:
	FILE	*file;
};

class WF_MPComponent;
class MultipartUploadData : public UploadData {
public:
	MultipartUploadData(const char *filename);
	~MultipartUploadData();
	int getSize() const;
	int send(NetIOStream *) const;
	const char *getContentType() const;

protected:
	char		*boundary;
	char		*sdata;
	FILE		*file;
	WF_MPComponent	*parts;
};

#endif /* __httpfetch_h */

